#define VY_DEFINE_DEPENDENCY_FILE_STRUCTURES
#include "asset_dependencies.h"

#include "aio.h"
#include "buffer_manager.h"

#include <assert.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

typedef struct {
    uint32_t begin;
    uint32_t count;
} vy_dep_list;

typedef struct {
    vy_uid *uids;
    vy_dep_list *lists;
    uint32_t capacity;
} vy_dep_map;

static vy_dep_map _map;
static vy_uid *_list_mem;

vy_result LoadAssetDependencies(void) {
    vy_dependency_file_header header;
    vy_file_id fid = vyAddFile("assets/deps.bin");

    if (vySubmitSingleLoadSync((vy_file_load){.dest      = &header,
                                              .num_bytes = sizeof(header),
                                              .offset    = 0,
                                              .file      = fid}) != VY_AIO_STATE_FINISHED) {
        vyReportError("core", "Failed to load deps.bin");
        return VY_UNKNOWN_ERROR;
    }
    void *buffer = vyAllocBuffer(header.data_size);
    if (vySubmitSingleLoadSync((vy_file_load){.dest      = buffer,
                                              .num_bytes = header.data_size,
                                              .offset    = sizeof(header),
                                              .file      = fid}) != VY_AIO_STATE_FINISHED) {
        vyReportError("core", "Failed to load deps.bin");
        return VY_UNKNOWN_ERROR;
    }

    /* We know the exact number of list entries */
    uint64_t total_list_entries =
        (header.data_size - header.num_lists * sizeof(vy_dependency_file_list_header)) /
        sizeof(vy_uid);
    _list_mem = malloc(total_list_entries * sizeof(vy_uid));
    if (!_list_mem) {
        vyReleaseBuffer(buffer, header.data_size);
        vyReportError("core", "Failed to allocate dependency list storage.");
        return VY_UNKNOWN_ERROR;
    }

    _map.capacity = vyNextPowerOfTwo32(header.num_lists);
    _map.uids     = calloc(_map.capacity, sizeof(vy_uid));
    if (!_map.uids) {
        free(_list_mem);
        vyReleaseBuffer(buffer, header.data_size);
        vyReportError("core", "Failed to allocate dependency list storage.");
        return VY_UNKNOWN_ERROR;
    }

    _map.lists = calloc(_map.capacity, sizeof(vy_dep_list));
    if (!_map.uids) {
        free(_list_mem);
        free(_map.uids);
        vyReleaseBuffer(buffer, header.data_size);
        vyReportError("core", "Failed to allocate dependency list storage.");
        return VY_UNKNOWN_ERROR;
    }

    uint32_t storage_at = 0;

    vy_dependency_file_list_header *list = buffer;
    for (uint32_t i = 0; i < header.num_lists; ++i) {
        const vy_uid *entries = (vy_uid *)(list + 1);

        /* Validate the checksum */
        XXH64_hash_t file_hash = XXH64_hashFromCanonical(&list->checksum);
        XXH64_hash_t calc_hash = XXH3_64bits(entries, sizeof(vy_uid) * list->num_entries);
        if (file_hash != calc_hash) {
            vyReportError("core", "Checksum mismatch in list %u", i);
            vyReleaseBuffer(buffer, header.data_size);
            return VY_UNKNOWN_ERROR;
        }

        /* Store the list */
        memcpy(_list_mem + storage_at, entries, sizeof(vy_uid) * list->num_entries);
        bool inserted = false;
        for (uint32_t j = 0; j < _map.capacity; ++j) {
            uint32_t slot = (list->uid + j) % _map.capacity;
            if (_map.uids[slot] == VY_INVALID_UID) {
                _map.uids[slot] = list->uid;
                _map.lists[slot].begin = storage_at;
                _map.lists[slot].count = list->num_entries;
                inserted = true;
                break;
            }
        }
        storage_at += list->num_entries;
        assert(inserted);
        assert(storage_at <= total_list_entries);

        list = (vy_dependency_file_list_header *)(entries + list->num_entries);
    }

    vyReleaseBuffer(buffer, header.data_size);

    return VY_SUCCESS;
}

void ReleaseAssetDependencies(void) {
    free(_list_mem);
    free(_map.uids);
    free(_map.lists);
}

VY_DLLEXPORT vy_asset_dependency_list vyGetAssetDependencies(vy_uid asset) {
    vy_asset_dependency_list result = {
        .dependencies = NULL,
        .count        = 0,
    };
    for (uint32_t i = 0; i < _map.capacity; ++i) {
        uint32_t slot = (asset + i) % _map.capacity;
        if (_map.uids[slot] == asset) {
            result.dependencies = &_list_mem[_map.lists[slot].begin];
            result.count        = _map.lists[slot].count;
            break;
        } else if (_map.uids[slot] == VY_INVALID_UID) {
            break;
        }
    }
    return result;
}