#include "assetmeta.h"
#include "utils.h"

#include "runtime/threading.h"
#include "runtime/aio.h"
#include "runtime/buffer_manager.h"

#include <xxhash/xxhash.h>
#include <stdio.h>

#define MAP_SIZE 2048
typedef struct {
    vy_file_id fids[MAP_SIZE];
    vy_uid uids[MAP_SIZE];
    vy_assetmeta meta[MAP_SIZE];

    unsigned int used_slots;
} vy_uid_map;

#pragma pack(push, 1)
typedef struct {
    XXH64_canonical_t checksum;
    uint32_t num_entries;
    uint32_t _reserved;
} vy_assetmeta_header;
#pragma pack(pop)

#pragma pack(push, 1)
typedef struct {
    vy_uid uid;
    vy_file_id source_file;
    vy_assetmeta meta;
} vy_assetmeta_entry;
#pragma pack(pop)

static vy_uid_map _map;
static vy_mutex *_guard;

vy_result vyLoadAssetMeta(void) {
    _guard = vyCreateMutex();
    
    /* Load the meta file */
    size_t fsz = vyGetFileSize("meta.bin");
    if (fsz == 0) {
        vyLog("ASSETC", "Metadata file 'meta.bin' not found. All assets will be processed.");
        return VY_SUCCESS;
    }

    void *buffer = vyAllocBuffer(fsz);
    if (!buffer) {
        vyReportError("ASSETC", "Failed to allocate buffer for holding asset metadata.");
        return 1;
    }
    
    vy_load_batch load;
    load.loads[0].dest = buffer;
    load.loads[0].file = vyAddFile("meta.bin");
    load.loads[0].num_bytes = fsz;
    load.loads[0].offset    = 0;
    load.num_loads          = 1;

    vy_aio_handle handle;
    if (vySubmitLoadBatch(&load, &handle) != VY_SUCCESS) {
        vyReportError("ASSETC", "Failed to submit load of 'meta.bin'.");
        vyReleaseBuffer(buffer, fsz);
        return 1;
    }
    if (vyWaitForAIOCompletion(handle) != VY_AIO_STATE_FINISHED) {
        vyReportError("ASSETC", "Failed to load 'meta.bin'.");
        vyReleaseBuffer(buffer, fsz);
        vyReleaseAIO(handle);
        return 1;
    }
    vyReleaseAIO(handle);

    const vy_assetmeta_header *header = buffer;
    const vy_assetmeta_entry *entries = (vy_assetmeta_entry *)(header + 1);
    
    if ((sizeof(vy_assetmeta_entry) * header->num_entries + sizeof(*header)) > fsz) {
        vyReportError("ASSETC", "Header of 'meta.bin' is corrupted: Mismatched num_entries and filesize.");
        vyReleaseBuffer(buffer, fsz);
        return 1;
    }
    XXH64_hash_t hash = XXH3_64bits(entries, sizeof(vy_assetmeta_entry) * header->num_entries);
    XXH64_hash_t header_hash = XXH64_hashFromCanonical(&header->checksum);
    if (hash != header_hash) {
        vyReportError("ASSETC",
                      "Metadata file 'meta.bin' is corrupted: Wrong checksum.");
        vyReleaseBuffer(buffer, fsz);
        return 1;
    }

    for (uint32_t i = 0; i < header->num_entries; ++i) {
        /* Load here to avoid unaligned pointer */
        vy_assetmeta meta = entries[i].meta;
        vyAddUIDMapping(entries[i].source_file, entries[i].uid, &meta);
    }

    vyReleaseBuffer(buffer, fsz);
    return 0;
}

vy_result vySaveAssetMeta(void) {
    vy_assetmeta_header header = {0};

    /* Count number of entries */
    for (size_t i = 0; i < MAP_SIZE; ++i) {
        if (_map.fids[i] != VY_INVALID_FILE_ID)
            header.num_entries += 1;
    }

    vy_assetmeta_entry *entries = vyAllocBuffer(sizeof(vy_assetmeta_entry) * header.num_entries);
    if (!entries)
        return VY_UNKNOWN_ERROR;

    /* Store entries */
    size_t j = 0;
    for (size_t i = 0; i < MAP_SIZE; ++i) {
        if (_map.fids[i] != VY_INVALID_FILE_ID) {
            entries[j].source_file = _map.fids[i];
            entries[j].uid         = _map.uids[i];
            entries[j].meta        = _map.meta[i];
            ++j;
        }
    }

    XXH64_hash_t hash = XXH3_64bits(entries, sizeof(vy_assetmeta_entry) * header.num_entries);
    XXH64_canonicalFromHash(&header.checksum, hash);
    header._reserved = 0;

    FILE *f = fopen("meta.bin", "wb");
    if (!f) {
        vyReportError("ASSETC", "Failed to open 'meta.bin'");
        vyReleaseBuffer(entries, sizeof(vy_assetmeta_entry) * header.num_entries);
        return VY_UNKNOWN_ERROR;
    }
    if (fwrite(&header, sizeof(header), 1, f) != 1) {
        vyReportError("ASSETC", "Failed to write 'meta.bin'");
        vyReleaseBuffer(entries, sizeof(vy_assetmeta_entry) * header.num_entries);
        return VY_UNKNOWN_ERROR;
    }
    if (fwrite(entries, sizeof(vy_assetmeta_entry), header.num_entries, f) != header.num_entries) {
        vyReportError("ASSETC", "Failed to write 'meta.bin'");
        vyReleaseBuffer(entries, sizeof(vy_assetmeta_entry) * header.num_entries);
        return VY_UNKNOWN_ERROR;
    }

    fclose(f);

    return VY_SUCCESS;
}

vy_uid vyLookupUID(vy_file_id fid) {
    vyLockMutex(_guard);
    unsigned int i = 0;
    vy_uid result  = VY_INVALID_UID;
    while (i < MAP_SIZE) {
        unsigned int slot = (fid + i) % MAP_SIZE;
        if (_map.fids[slot] == fid) {
            result          = _map.uids[slot];
            break;
        } else if (_map.fids[slot] == VY_INVALID_FILE_ID) {
            break;
        }
        ++i;
    }
    vyUnlockMutex(_guard);
    return result;
}

bool vyGetAssetMeta(vy_file_id source_file, vy_assetmeta *meta) {
    vyLockMutex(_guard);
    unsigned int i = 0;
    bool result    = false;
    while (i < MAP_SIZE) {
        unsigned int slot = (source_file + i) % MAP_SIZE;
        if (_map.fids[slot] == source_file) {
            *meta  = _map.meta[slot];
            result = true;
            break;
        } else if (_map.fids[slot] == VY_INVALID_FILE_ID) {
            break;
        }
        ++i;
    }
    vyUnlockMutex(_guard);
    return result;
}

void vyAddUIDMapping(vy_file_id fid, vy_uid uid, const vy_assetmeta *meta) {
    vyLockMutex(_guard);
    float fill_rate = (float)_map.used_slots / MAP_SIZE;
    if (fill_rate >= .5f) {
        vyLog("ASSETC", "UID map is above 50% filled.");
    }
    unsigned int i = 0;
    while (i < MAP_SIZE) {
        unsigned int slot = (fid + i) % MAP_SIZE;
        if (_map.fids[slot] == VY_INVALID_FILE_ID) {
            _map.fids[slot] = fid;
            _map.uids[slot] = uid;
            if (meta) {
                _map.meta[slot] = *meta;
            } else {
                _map.meta[slot].last_changed = 0;
                _map.meta[slot].compiled_ts = 0;
                _map.meta[slot].processing_flags = 0;
            }
            ++_map.used_slots;
            break;
        } else if (_map.fids[slot] == fid) {
            break;
        }
        ++i;
    }
    if (i == MAP_SIZE) {
        vyReportError("ASSETC", "Failed to insert entry into UID map.");
    }
    vyUnlockMutex(_guard);
}